/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *  (C) 2008 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "hydra.h"
#include "hydra_utils.h"
#include "mpiexec.h"
#include "lchu.h"
#include "csi.h"

HYD_Handle handle;

static void usage(void)
{
    printf("\n");
    printf("Usage: ./mpiexec [global opts] [exec1 local opts] : [exec2 local opts] : ...\n\n");

    printf("Global Options (passed to all executables):\n");
    printf("\t--verbose                        [Verbose mode]\n");
    printf("\t--version                        [Version information]\n");
    printf("\t--enable-x/--disable-x           [Enable or disable X forwarding]\n");
    printf("\t--proxy-port                     [Port on which proxies can listen]\n");
    printf("\t--bootstrap                      [Bootstrap server to use]\n");
    printf("\t-genv {name} {value}             [Environment variable name and value]\n");
    printf("\t-genvlist {env1,env2,...}        [Environment variable list to pass]\n");
    printf("\t-genvnone                        [Do not pass any environment variables]\n");
    printf("\t-genvall                         [Pass all environment variables (default)]\n");
    printf("\t-f {name}                        [File containing the host names]\n");
    printf("\t-wdir {dirname}                  [Working directory to use]\n");

    printf("\n");

    printf("Local Options (passed to individual executables):\n");
    printf("\t-n/-np {value}                   [Number of processes]\n");
    printf("\t-env {name} {value}              [Environment variable name and value]\n");
    printf("\t-envlist {env1,env2,...}         [Environment variable list to pass]\n");
    printf("\t-envnone                         [Do not pass any environment variables]\n");
    printf("\t-envall                          [Pass all environment variables (default)]\n");
    printf
        ("\t{exec_name} {args}               [Name of the executable to run and its arguments]\n");

    printf("\n");
}


int main(int argc, char **argv)
{
    struct HYD_Proc_params *proc_params;
    struct HYD_Partition_list *partition;
    int exit_status = 0;
    int timeout;
    HYD_Status status = HYD_SUCCESS;

    HYDU_FUNC_ENTER();

    status = HYD_LCHI_Get_parameters(argc, argv);
    if (status == HYD_GRACEFUL_ABORT) {
        exit(0);
    }
    else if (status != HYD_SUCCESS) {
        usage();
        goto fn_fail;
    }

    if (handle.debug) {
        status = HYD_LCHI_Print_parameters();
        HYDU_ERR_POP(status, NULL);
    }

    /* Convert the host file to a host list */
    status = HYD_LCHU_Create_host_list();
    HYDU_ERR_POP(status, "unable to create host list\n");

    /* Consolidate the environment list that we need to propagate */
    status = HYD_LCHU_Create_env_list();
    HYDU_ERR_POP(status, "unable to create env list\n");

    proc_params = handle.proc_params;
    while (proc_params) {
        proc_params->stdout_cb = HYD_LCHI_stdout_cb;
        proc_params->stderr_cb = HYD_LCHI_stderr_cb;
        proc_params = proc_params->next;
    }
    handle.stdin_cb = HYD_LCHI_stdin_cb;

    HYDU_Time_set(&handle.start, NULL); /* NULL implies right now */
    if (getenv("MPIEXEC_TIMEOUT"))
        timeout = atoi(getenv("MPIEXEC_TIMEOUT"));
    else
        timeout = -1;   /* Set a negative timeout */
    HYDU_Debug("Timeout set to %d seconds (negative means infinite)\n", timeout);
    HYDU_Time_set(&handle.timeout, &timeout);

    /* Launch the processes */
    status = HYD_CSI_Launch_procs();
    HYDU_ERR_POP(status, "control system error launching processes\n");

    /* Wait for their completion */
    status = HYD_CSI_Wait_for_completion();
    HYDU_ERR_POP(status, "control system error waiting for completion\n");

    /* Check for the exit status for all the processes */
    exit_status = 0;
    for (proc_params = handle.proc_params; proc_params; proc_params = proc_params->next)
        for (partition = proc_params->partition; partition; partition = partition->next)
            exit_status |= partition->exit_status;

    /* Call finalize functions for lower layers to cleanup their resources */
    status = HYD_CSI_Finalize();
    HYDU_ERR_POP(status, "control system error on finalize\n");

    /* Free the mpiexec params */
    HYD_LCHU_Free_params();

  fn_exit:
    HYDU_FUNC_EXIT();
    if (status == HYD_GRACEFUL_ABORT)
        return 0;
    else if (status != HYD_SUCCESS)
        return -1;
    else
        return exit_status;

  fn_fail:
    goto fn_exit;
}
